/* an example of calling the NAG Sobol generator nag_quasi_rand_uniform (g05ymc)
   in an OpenMP multi-threaded environment */
#include <stdio.h>
#include <math.h>
#include <nag.h>
#include <nagx04.h>
#include <nag_stdlib.h>
#include <nagg05.h>
#include <omp.h>

int main(int argc, char *argv[]) {
  int i, exit_status = 0;
  Integer NumberOfThreads, liref, pdquas, iskip, tskip, ThreadNo, npaths, 
    idim, tnpaths, t1;
  double sumRandom;
  Integer *iref = 0;
  double *quas = 0;  
  NagError fail;

  /* we are currently just summing the values for the first dimension, so makes
     sense to work in ColMajor order, where all values from a given dimension 
     are contiguous */
  Nag_OrderType order = Nag_ColMajor;

  /* want the Sobol generator */
  Nag_QuasiRandom_Sequence genid = Nag_QuasiRandom_Sobol;

  /* parse the input arguments */
  if (argc != 4) {
    printf("Usage: XXX n d s\n");
    printf("Where:\n");
    printf("  n = number of paths\n");
    printf("  d = number of dimensions\n");
    printf("  s = initial skip\n");
    return 0;
  }

  /* get the problem size */
  npaths = atoi(argv[1]);
  idim = atoi(argv[2]);
  iskip = atoi(argv[3]);

  /* array sizes */
  liref = 32 * idim + 7;

#pragma omp parallel default(none) \
  reduction(+:sumRandom) \
    private(ThreadNo,tskip,tnpaths,quas,iref,fail,i,t1,pdquas,NumberOfThreads) \
      shared(npaths,exit_status,idim,liref,iskip,genid,order)
  {

    /* intialise the error structure */
    INIT_FAIL(fail);

    /* thread number */
    ThreadNo = omp_get_thread_num();

    /* number of threads */
    NumberOfThreads = omp_get_num_threads();

    /* calculate the skip and the number of values to generate for this thread */
    if (NumberOfThreads == 1) {
      tskip = 0;
      tnpaths = npaths;
    } else {
      t1 = npaths % NumberOfThreads;
      t1 = (ThreadNo < t1) ? ThreadNo : t1;
      tskip = ThreadNo * (npaths / NumberOfThreads) + t1;
      tnpaths = (npaths + NumberOfThreads - ThreadNo - 1) / NumberOfThreads;
    } 

    /* adjust the skip for this thread by the initial skip */
    tskip += iskip;

    /* calculate the stride */
    pdquas = (order == Nag_RowMajor)?idim:tnpaths;

    /* allocate memory to hold the generated values (quas) and state of the 
       quasi-random generator (iref) */
    if (!(quas = NAG_ALLOC(tnpaths*idim, double)) ||
        !(iref = NAG_ALLOC(liref, Integer)))
      {
        printf("Allocation failure in thread %d\n", ThreadNo + 1);
        exit_status = -1;
        goto TEND;
      }
    
    /* nag_quasi_init (g05ylc): initialise the quasi generator */
    nag_quasi_init(genid, idim, iref, liref, tskip, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_quasi_init (g05ylc) in thread %d.\n%s\n", 
               ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }

    /* nag_quasi_rand_uniform (g05ymc): generate the required tnpaths values */
    nag_quasi_rand_uniform(order, tnpaths, quas, pdquas, iref, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_quasi_rand_uniform (g05ymc) in thread %d.\n%s\n",
               ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }
    
    /* calculate the sum for this particular thread. We are only interested in
       the first dimension which are stored in the first TNPATHS of the QUAS 
       array */
    for (i = 0; i < tnpaths; i++)
      sumRandom += quas[i];

    printf("Thread = %4"NAG_IFMT", Sum = %g\n", ThreadNo+1,sumRandom);

    /* free memory */
  TEND:
    if (quas) free(quas);
    if (iref) free(iref);
  }
  
  /* display the results */
  if (!exit_status) {
    printf("\nTotal Sum = %g\n", sumRandom);
  }

  return exit_status;
}
